@@ -6,10 +6,228 @@
 #include <adelie_core/util/macros.hpp>
 #include <adelie_core/util/stopwatch.hpp>
 #include <adelie_core/util/tqdm.hpp>
+#if defined(_OPENMP)
+#include <omp.h>
+#endif
 
 namespace adelie_core {
 namespace solver {
 
+/**
+ * Updates absolute gradient in the base state.
+ * The state DOES NOT have to be in its invariance. 
+ * After the function finishes, abs_grad will reflect the correct value
+ * respective to grad.
+ */
+template <class StateType, class ValueType>
+inline void update_abs_grad(
+    StateType& state,
+    ValueType lmda
+)
+{
+    using state_t = std::decay_t<StateType>;
+    using vec_value_t = typename state_t::vec_value_t;
+    using rowmat_uint64_t = util::rowmat_type<uint64_t>;
+
+    const auto& constraints = state.constraints;
+    const auto& groups = state.groups;
+    const auto& group_sizes = state.group_sizes;
+    const auto& penalty = state.penalty;
+    const auto& grad = state.grad;
+    const auto& screen_set = state.screen_set;
+    const auto& screen_hashset = state.screen_hashset;
+    const auto& screen_begins = state.screen_begins;
+    const auto& screen_beta = state.screen_beta;
+    const auto alpha = state.alpha;
+    const auto constraint_buffer_size = state.constraint_buffer_size;
+    const auto n_threads = state.n_threads;
+    auto& abs_grad = state.abs_grad;
+
+    const auto is_screen = [&](auto i) {
+        return screen_hashset.find(i) != screen_hashset.end();
+    };
+
+    vec_value_t buff(group_sizes.maxCoeff());
+    rowmat_uint64_t constraint_buffer(std::max<size_t>(1, n_threads), constraint_buffer_size);
+
+    // do not parallelize since it may result in large false sharing 
+    // (access to abs_grad[i] is random order)
+    for (size_t ss_idx = 0; ss_idx < screen_set.size(); ++ss_idx) 
+    {
+        const auto i = screen_set[ss_idx];
+        const auto b = screen_begins[ss_idx]; 
+        const auto k = groups[i];
+        const auto size_k = group_sizes[i];
+        const auto pk = penalty[i];
+        const auto regul = ((1-alpha) * lmda) * pk;
+        const auto constraint = constraints[i];
+        const Eigen::Map<const vec_value_t> sbeta(
+            screen_beta.data() + b,
+            size_k
+        );
+        const auto common_expr = grad.segment(k, size_k) - regul * sbeta;
+
+        if (constraint == nullptr) {
+            abs_grad[i] = common_expr.matrix().norm();
+        } else {
+            auto vbuff = buff.head(size_k);
+            constraint->gradient(sbeta, vbuff);
+            abs_grad[i] = (common_expr - vbuff).matrix().norm();
+        }
+    }
+
+    // can be parallelized since access is in linear order.
+    // any false sharing is happening near the beginning/ends of the block of indices.
+    std::atomic_bool try_failed = false; 
+    const auto routine = [&](int i) {
+        if (try_failed.load(std::memory_order_relaxed) || is_screen(i)) return;
+        #if defined(_OPENMP)
+        auto cbuff = constraint_buffer.row(omp_get_thread_num());
+        #else
+        auto cbuff = constraint_buffer.row(0);
+        #endif
+        const auto k = groups[i];
+        const auto size_k = group_sizes[i];
+        const auto constraint = constraints[i];
+        const auto v_k = grad.segment(k, size_k);
+        try {
+            abs_grad[i] = (
+                constraint ?
+                constraint->solve_zero(v_k, cbuff) :
+                v_k.matrix().norm()
+            );
+        } catch (...) {
+            try_failed = true;
+        }
+    };
+    if (n_threads <= 1) {
+        for (int i = 0; i < groups.size(); ++i) routine(i);
+    } else {
+        #if defined(_OPENMP)
+        #pragma omp parallel for schedule(static) num_threads(n_threads)
+        #endif
+        for (int i = 0; i < groups.size(); ++i) routine(i);
+    }
+    if (try_failed) {
+        throw util::adelie_core_solver_error(
+            "exception raised in constraint->solve_zero(). "
+            "Try changing the configurations such as convergence tolerance that affect solve_zero(). "
+        );
+    }
+}
+
+/**
+ * Updates all derived quantities from screen_set in the base class. 
+ * The state must be such that only the screen_set is either unchanged from invariance,
+ * or appended with new groups.
+ * After the function finishes, all screen quantities in the base class
+ * will be consistent with screen_set, and the state is otherwise effectively
+ * unchanged in the sense that other quantities dependent on screen states are unchanged.
+ */
+template <class StateType>
+inline void update_screen_derived_base(
+    StateType& state
+)
+{
+    const auto& group_sizes = state.group_sizes;
+    const auto& screen_set = state.screen_set;
+    auto& screen_hashset = state.screen_hashset;
+    auto& screen_begins = state.screen_begins;
+    auto& screen_beta = state.screen_beta;
+    auto& screen_is_active = state.screen_is_active;
+
+    /* update screen_hashset */
+    const auto old_screen_size = screen_begins.size();
+    const auto screen_set_new_begin = std::next(screen_set.begin(), old_screen_size);
+    screen_hashset.insert(screen_set_new_begin, screen_set.end());
+
+    /* update screen_begins */
+    size_t screen_value_size = (
+        (old_screen_size == 0) ? 
+        0 : (screen_begins.back() + group_sizes[screen_set[old_screen_size-1]])
+    );
+    for (size_t i = old_screen_size; i < screen_set.size(); ++i) {
+        const auto curr_size = group_sizes[screen_set[i]];
+        screen_begins.push_back(screen_value_size);
+        screen_value_size += curr_size;
+    }
+
+    /* update screen_beta */
+    screen_beta.resize(screen_value_size, 0);
+
+    /* update screen_is_active */
+    screen_is_active.resize(screen_set.size(), false);
+}
+
+// TODO: 
+// It might be too much to save all of them though especially if we have a lot of active constraints.
+// Then, state.dual is going to be super dense and we have 100 rows of such dense vectors.
+// Nonetheless, it's still useful to have this for diagnostic purposes... maybe a flag to control whether this gets saved?
+template <class StateType, class VecIndexType, class VecValueType>
+inline auto sparsify_dual(
+    const StateType& state,
+    VecIndexType& indices,
+    VecValueType& values
+)
+{
+    using vec_index_t = typename StateType::vec_index_t;
+    using vec_value_t = typename StateType::vec_value_t;
+    using sp_vec_value_t = typename StateType::sp_vec_value_t;
+
+    const auto& constraints = state.constraints;
+    const auto& dual_groups = state.dual_groups;
+    const auto n_threads = state.n_threads;
+
+    const auto n_constraints = constraints.size();
+    vec_index_t begins(n_constraints+1);
+    begins[0] = 0;
+    begins.tail(n_constraints) = vec_index_t::NullaryExpr(
+        n_constraints, 
+        [&](auto i) {
+            const auto constraint = constraints[i];
+            return constraint ? constraint->duals_nnz() : 0;
+        }
+    );
+    for (Eigen::Index i = 1; i < begins.size(); ++i) {
+        begins[i] += begins[i-1];
+    }
+    indices.resize(begins[n_constraints]); 
+    values.resize(begins[n_constraints]);
+
+    if (begins[n_constraints]) {
+        const auto routine = [&](auto i) {
+            const auto b = begins[i];
+            const auto nnz = begins[i+1] - b;
+            if (nnz <= 0) return;
+            const auto constraint = constraints[i];
+            Eigen::Map<vec_index_t> indices_v(indices.data() + b, nnz);
+            Eigen::Map<vec_value_t> values_v(values.data() + b, nnz);
+            constraint->dual(indices_v, values_v);
+            indices_v += dual_groups[i];
+        };
+        
+        if (n_threads <= 1) {
+            for (Eigen::Index i = 0; i < static_cast<Eigen::Index>(n_constraints); ++i) routine(i);
+        } else {
+            #pragma omp parallel for schedule(static) num_threads(n_threads)
+            for (Eigen::Index i = 0; i < static_cast<Eigen::Index>(n_constraints); ++i) routine(i);
+        }
+    }
+
+    const auto last_constraint = constraints[n_constraints-1];
+    const auto n_duals = (
+        dual_groups[n_constraints-1] + 
+        (last_constraint ? last_constraint->duals() : 0)
+    );
+    Eigen::Map<const sp_vec_value_t> dual_map(
+        n_duals,
+        indices.size(),
+        indices.data(),
+        values.data()
+    );
+    return dual_map;
+}
+
 template <class StateType, class PBType>
 ADELIE_CORE_STRONG_INLINE
 void pb_add_suffix(
@@ -59,8 +277,7 @@ bool early_exit(
  * All derived screen quantities must be updated afterwards. 
  */
 template <class StateType, class ValueType>
-ADELIE_CORE_STRONG_INLINE 
-void screen(
+inline void screen(
     StateType& state,
     ValueType lmda_next,
     bool all_kkt_passed,
@@ -221,15 +438,17 @@ bool kkt(
     return true;
 }
 
-template <class StateType,
-          class PBType,
-          class PBAddSuffixType,
-          class UpdateLossNullType,
-          class UpdateInvarianceType,
-          class UpdateSolutionsType,
-          class EarlyExitType,
-          class ScreenType,
-          class FitType>
+template <
+    class StateType,
+    class PBType,
+    class PBAddSuffixType,
+    class UpdateLossNullType,
+    class UpdateInvarianceType,
+    class UpdateSolutionsType,
+    class EarlyExitType,
+    class ScreenType,
+    class FitType
+>
 inline void solve_core(
     StateType&& state,
     PBType&& pb,
@@ -305,7 +524,7 @@ inline void solve_core(
     // Generate lambda path if needed
     // ==================================================================================== 
     if (setup_lmda_path) {
-        if (lmda_path_size <= 0) throw util::adelie_core_solver_error("lmda_path_size must be > 0.");
+        if (lmda_path_size <= 0) return;
 
         lmda_path.resize(lmda_path_size);
 
@@ -349,14 +568,6 @@ inline void solve_core(
             if (i < large_lmda_path.size()-1) {
                 // update progress bar
                 static_cast<void>(pb_it != pb.end());
-
-                // Do not check for early-stopping.
-                // This choice is deliberate for these reasons:
-                // We are in this loop only because the user has supplied their path.
-                // In this case, we should provide what they asked for.
-                // If alpha == 0, this loop is costly but this is also the only time the solutions are actually changing.
-                // If alpha > 0, the solution is the same at every iteration.
-                // Depending on alpha, we have different reasons for not early stopping.
             }
 
             auto tup = fit_f(state, large_lmda_path[i]);
@@ -372,6 +583,12 @@ inline void solve_core(
                 );
                 pb_add_suffix_f(state, pb);
                 ++pb_it;
+
+                if (early_exit_f(state)) {
+                    if (!(pb_it != pb.end())) return;
+                    pb_add_suffix_f(state, pb);
+                    return;
+                }
             // otherwise, put the state at the last fitted lambda (lmda_max)
             } else {
                 update_invariance_f(state, state_gaussian_pin, large_lmda_path[i]);
@@ -393,12 +610,6 @@ inline void solve_core(
 
     for (; pb_it != pb.end(); ++pb_it)
     {
-        // check early exit
-        if (early_exit_f(state)) {
-            pb_add_suffix_f(state, pb);
-            break;
-        }
-
         // batch the next set of lambdas
         const auto lmda_curr = lmda_path[lmda_path_idx];
 
@@ -468,7 +679,17 @@ inline void solve_core(
         } // end while(1)
 
         pb_add_suffix_f(state, pb);
-    }
+
+        // Early exit condition must be here to ensure that 
+        // it is called after processing each lambda (including the last lambda).
+        if (early_exit_f(state)) {
+            // must add one more bar
+            ++pb_it;
+            if (!(pb_it != pb.end())) break;
+            pb_add_suffix_f(state, pb);
+            break;
+        }
+    } // end for-loop over progress bar
 }
 
 } // namespace solver
