@@ -1,9 +1,8 @@
 #pragma once
 #include <adelie_core/configs.hpp>
-#include <adelie_core/solver/solver_base.hpp>
+#include <adelie_core/solver/solver_gaussian_naive.hpp>
 #include <adelie_core/solver/solver_gaussian_pin_naive.hpp>
 #include <adelie_core/state/state_gaussian_pin_naive.hpp>
-#include <adelie_core/state/state_glm_naive.hpp>
 
 namespace adelie_core {
 namespace solver {
@@ -50,16 +49,76 @@ struct GlmNaiveBufferPack
 
     dyn_vec_value_t screen_beta_prev;
     dyn_vec_bool_t screen_is_active_prev;
-    dyn_vec_value_t screen_dual_prev;
 
     vec_value_t ones;       // (n,) vector of ones
     vec_value_t buffer_n;   // (n,) extra buffer
 };
 
-template <class StateType, 
-          class GlmType,
-          class StateGaussianPinType,
-          class ValueType>
+/**
+ * Unlike the similar function in gaussian::naive,
+ * this does not call the base version to update the base classes's screen derived quantities.
+ * This is because in GLM fitting, the three screen_* inputs are modified at every IRLS loop,
+ * while the base quantities remain the same. 
+ * It is only when IRLS finishes and we must screen for variables where we have to update the base quantities.
+ * In gaussian naive setting, the IRLS has loop size of 1 essentially, so the two versions are synonymous.
+ */
+template <
+    class StateType, 
+    class XMType, 
+    class WType,
+    class SXMType, 
+    class STType, 
+    class SVType
+>
+inline void update_screen_derived(
+    StateType& state,
+    const XMType& X_means,
+    const WType& weights_sqrt,
+    size_t begin,
+    size_t size,
+    SXMType& screen_X_means,
+    STType& screen_transforms,
+    SVType& screen_vars
+)
+{
+    const auto& group_sizes = state.group_sizes;
+    const auto& screen_set = state.screen_set;
+    const auto& screen_begins = state.screen_begins;
+
+    const auto new_screen_size = screen_set.size();
+    const int new_screen_value_size = (
+        (screen_begins.size() == 0) ? 0 : (
+            screen_begins.back() + group_sizes[screen_set.back()]
+        )
+    );
+
+    screen_X_means.resize(new_screen_value_size);    
+    screen_transforms.resize(new_screen_size);
+    screen_vars.resize(new_screen_value_size, 0);
+
+    gaussian::naive::update_screen_derived(
+        *state.X,
+        X_means,
+        weights_sqrt,
+        state.groups,
+        state.group_sizes,
+        state.screen_set,
+        state.screen_begins,
+        begin,
+        size,
+        state.intercept,
+        screen_X_means,
+        screen_transforms,
+        screen_vars
+    );
+}
+
+template <
+    class StateType, 
+    class GlmType,
+    class StateGaussianPinType,
+    class ValueType
+>
 ADELIE_CORE_STRONG_INLINE
 void update_solutions(
     StateType& state,
@@ -68,15 +127,26 @@ void update_solutions(
     ValueType lmda
 )
 {
+    using state_t = std::decay_t<StateType>;
+    using vec_index_t = typename state_t::vec_index_t;
+    using vec_value_t = typename state_t::vec_value_t;
+    using sp_vec_value_t = typename state_t::sp_vec_value_t;
+
     const auto loss_null = state.loss_null;
     const auto loss_full = state.loss_full;
     const auto& eta = state.eta;
     auto& betas = state.betas;
+    auto& duals = state.duals;
     auto& devs = state.devs;
     auto& lmdas = state.lmdas;
     auto& intercepts = state.intercepts;
 
+    vec_index_t dual_indices; 
+    vec_value_t dual_values;
+
     betas.emplace_back(std::move(state_gaussian_pin_naive.betas.back()));
+    sp_vec_value_t dual = sparsify_dual(state, dual_indices, dual_values);
+    duals.emplace_back(std::move(dual));
     intercepts.emplace_back(state_gaussian_pin_naive.intercepts.back());
     lmdas.emplace_back(lmda);
 
@@ -87,9 +157,11 @@ void update_solutions(
     );
 }
 
-template <class StateType,
-          class GlmType,
-          class BufferPackType>
+template <
+    class StateType,
+    class GlmType,
+    class BufferPackType
+>
 ADELIE_CORE_STRONG_INLINE
 void update_loss_null(
     StateType& state,
@@ -159,13 +231,14 @@ void update_loss_null(
     }
 }
 
-template <class StateType,
-          class GlmType,
-          class BufferPackType,
-          class ValueType,
-          class CUIType=util::no_op>
-ADELIE_CORE_STRONG_INLINE
-auto fit(
+template <
+    class StateType,
+    class GlmType,
+    class BufferPackType,
+    class ValueType,
+    class CUIType=util::no_op
+>
+inline auto fit(
     StateType& state,
     GlmType& glm,
     BufferPackType& buffer_pack,
@@ -199,7 +272,7 @@ auto fit(
     const auto& offsets = state.offsets;
     const auto& screen_set = state.screen_set;
     const auto& screen_begins = state.screen_begins;
-    const auto& screen_dual_begins = state.screen_dual_begins;
+    const auto constraint_buffer_size = state.constraint_buffer_size;
     const auto intercept = state.intercept;
     const auto max_active_size = state.max_active_size;
     const auto irls_max_iters = state.irls_max_iters;
@@ -213,7 +286,6 @@ auto fit(
     const auto loss_full = state.loss_full;
     auto& screen_beta = state.screen_beta;
     auto& screen_is_active = state.screen_is_active;
-    auto& screen_dual = state.screen_dual;
     auto& active_set_size = state.active_set_size;
     auto& active_set = state.active_set;
     auto& beta0 = state.beta0;
@@ -234,7 +306,6 @@ auto fit(
     auto& hess = buffer_pack.hess;
     auto& screen_beta_prev = buffer_pack.screen_beta_prev;
     auto& screen_is_active_prev = buffer_pack.screen_is_active_prev;
-    auto& screen_dual_prev = buffer_pack.screen_dual_prev;
 
     util::rowvec_type<value_t, 1> lmda_path_adjusted;
 
@@ -244,12 +315,10 @@ auto fit(
     const auto save_prev_valid = [&]() {
         screen_beta_prev = screen_beta;
         screen_is_active_prev = screen_is_active;
-        screen_dual_prev = screen_dual;
     };
     const auto load_prev_valid = [&]() {
         screen_beta.swap(screen_beta_prev);
         screen_is_active.swap(screen_is_active_prev);
-        screen_dual.swap(screen_dual_prev);
     };
 
     value_t screen_time = 0;
@@ -301,7 +370,7 @@ auto fit(
         }
         // this call should only adjust the size of screen_* quantities
         // and repopulate every entry using the new weights.
-        state::glm::naive::update_screen_derived(
+        update_screen_derived(
             state,
             X_means,
             irls_weights_sqrt,
@@ -329,12 +398,10 @@ auto fit(
             Eigen::Map<const vec_value_t>(screen_vars.data(), screen_vars.size()), 
             Eigen::Map<const vec_value_t>(screen_X_means.data(), screen_X_means.size()), 
             screen_transforms,
-            Eigen::Map<const vec_index_t>(screen_dual_begins.data(), screen_dual_begins.size()),
             lmda_path_adjusted,
+            constraint_buffer_size,
             intercept, max_active_size, max_iters, 
-            // TODO: still unclear whether we should be max'ing or not.
-            // tolerance is relative to the scaling of null deviance and current total weight sum
-            tol * std::max<value_t>((loss_null - loss_full) / hess_sum, 1), 
+            tol * (loss_null - loss_full) / hess_sum, 
             0 /* adev_tol */, 0 /* ddev_tol */,
             newton_tol, newton_max_iters, n_threads,
             0 /* rsq (no need to track) */,
@@ -342,15 +409,11 @@ auto fit(
             resid_sum,
             Eigen::Map<vec_value_t>(screen_beta.data(), screen_beta.size()), 
             Eigen::Map<vec_safe_bool_t>(screen_is_active.data(), screen_is_active.size()),
-            Eigen::Map<vec_value_t>(screen_dual.data(), screen_dual.size()),
             active_set_size,
             active_set
         );
         try { 
-            gaussian::pin::naive::solve(
-                state_gaussian_pin_naive, 
-                check_user_interrupt
-            );
+            state_gaussian_pin_naive.solve(check_user_interrupt);
         } catch(...) {
             load_prev_valid();
             throw;
@@ -380,16 +443,7 @@ auto fit(
         glm.gradient(eta, resid); 
 
         /* check convergence */
-        // check directional derivative of gradient (resid) as an approximation
-        // to the quadratic loss. 
-        const auto& active_set = state_gaussian_pin_naive.active_set;
-        const auto& active_begins = state_gaussian_pin_naive.active_begins;
-        const auto n_active = (
-            (active_begins.size() == 0) ? 1 : (
-                active_begins.back() + group_sizes[screen_set[active_set[active_set_size-1]]]
-            )
-        );
-        if (std::abs(((resid - resid_prev) * (eta - eta_prev)).sum()) <= irls_tol * n_active) {
+        if (std::abs(((resid - resid_prev) * (eta - eta_prev)).sum()) <= irls_tol) {
             return std::make_tuple(
                 std::move(state_gaussian_pin_naive),
                 screen_time,
@@ -401,13 +455,15 @@ auto fit(
     }
 }
 
-template <class StateType,
-          class GlmType,
-          class PBType,
-          class ExitCondType,
-          class UpdateLossNullType,
-          class TidyType,
-          class CUIType>
+template <
+    class StateType,
+    class GlmType,
+    class PBType,
+    class ExitCondType,
+    class UpdateLossNullType,
+    class TidyType,
+    class CUIType
+>
 inline void solve(
     StateType&& state,
     GlmType&& glm,
@@ -440,7 +496,7 @@ inline void solve(
         const auto& ones = buffer_pack.ones;
         state.lmda = lmda;
         X.mul(resid, ones, grad);
-        state::update_abs_grad(state, lmda);
+        update_abs_grad(state, lmda);
     };
     const auto update_solutions_f = [&](auto& state, auto& state_gaussian_pin_naive, auto lmda) {
         update_solutions(
@@ -461,7 +517,7 @@ inline void solve(
             kkt_passed,
             n_new_active
         );
-        state::update_screen_derived_base(state);
+        update_screen_derived_base(state);
     };
     const auto fit_f = [&](auto& state, auto lmda) {
         return fit(
@@ -486,11 +542,13 @@ inline void solve(
     );
 }
 
-template <class StateType,
-          class GlmType,
-          class PBType,
-          class ExitCondType,
-          class CUIType=util::no_op>
+template <
+    class StateType,
+    class GlmType,
+    class PBType,
+    class ExitCondType,
+    class CUIType=util::no_op
+>
 inline void solve(
     StateType&& state,
     GlmType&& glm,
